import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.Scanner;

import cern.jet.random.*;
import cern.jet.random.engine.DRand;
import cern.jet.random.engine.RandomEngine;

/**
 * Synthetic data generation for SelfJoin application. We assume
 * 
 * N = Number of items = 20
 * L = Number of maximal potentially large itemsets = 40 (~ double the number of items)
 * D = Number of transactions = 100
 * T = Average size of transaction = 5
 * I = Average size of the maximal potentially large itemsets = 3
 * B = Average length of an item in bytes ( including payload per item such as item#, storeid etc.) = 20 
 * 
 * There are three steps of data generation:
 * 1. creating items : count = N, item-size = B bytes
 * 		we assume that each item is of fixed size = B bytes.
 * 		random number generation using letters [a-z]
 * 2. creating large itemsets of size L
 * 3. creating transactions of size D
 * 
 * The size of Output file of transactions is determined by: T * B * D 
 * 
 * Ref: R. Agrawal and R. Srikant. Fast algorithms for mining association
rules. In Proc. of 20th VLDB, 1215:487499, 1994.
 * 
 * NOTE: This file requires Colt library to be installed http://acs.lbl.gov/software/colt/install.html
 * @author Faraz Ahmad
 *
 */

public class CreateSelfJoinData {

	/**
	 * @param args
	 * Usage: CreateSelfJoinData [-out<i>outputdir</i>][-N][-L][-D][-T][-I][-B]
	 */
	private static int N = 20000;
	private static int B = 20;
	private static int L = 40000; 
	private static int D = 1500000000;
	private static int I = 10;
	private static int T = 20; // should be ~ 1.5 - 3 X I
	private static double correlation = 0.5; // used in item sets generation
	private static int MAX = 10;
	private static float HALF = 0.5f;
	private static String comma = ",";
	private static String newline = "\n";

	private static File dataPath = new File("/home/hadoop1/input_data/SelfJoinData");
	private static File itemsFile = new File(dataPath + "/items.txt");
	private static File itemSetsFile = new File(dataPath + "/itemsets.txt");
	private static File transactionsFile = new File(dataPath + "/transactions_new1.txt");
	
	
	public static void createItems() throws IOException {
		System.out.println("Step 1 of 3: Items generation");
		System.out.println("=============================");
		System.out.println("Number of items = " + N);
		System.out.println("Item size = " + B + " bytes");
		
		
		String entry = new String("");

		if(!itemsFile.exists()){
			itemsFile.createNewFile();
		}
		else {
			System.out.println("File " + itemsFile.getAbsolutePath() + " exists, skipping Item Generation step");	
			return;
		}
		OutputStream out = new FileOutputStream(itemsFile);
    
		Random s;
		int limit = 9;
		int num[] = new int[B-9]; // leaving 8 bytes for 'entryNum and one byte for '\n'
		byte[] buf = new byte[B];
		System.out.println("Writing to file: " + itemsFile.getAbsolutePath());

		for(int k = 0; k < N; k++){
			entry = new String("entryNum");
			s = new Random();
			for (int i = 0; i < num.length; i++){
				num[i] = s.nextInt(limit);
				entry = entry + num[i];
			}
			entry = entry + "\n";
			buf = entry.getBytes();
			out.write(buf, 0, buf.length);
		}
		out.close();

	}

	public static void createItemSets() throws IOException {
		System.out.println("Step 2 of 3: ItemSets generation");
		System.out.println("=============================");
		System.out.println("Number of large sets = " + L);
		
		if(!itemSetsFile.exists()){
			itemSetsFile.createNewFile();
		}
		else {
			System.out.println("File " + itemSetsFile.getAbsolutePath() + " exists, skipping LargeSets generation step");	
			return;
		}
		
		Scanner scanner = new Scanner(itemsFile);
		
		String[] items = new String[N];
		
		int i = 0;
		while (scanner.hasNext()){
			items[i++] = scanner.next();
		}
		
		OutputStream out = new FileOutputStream(itemSetsFile);
		System.out.println("Writing to file: " + itemSetsFile.getAbsolutePath());

		RandomEngine engine = new DRand();
		double mean = (double) I;
		Poisson poisson = new Poisson(mean, engine);
		Exponential exp = new Exponential(correlation, engine);
		Random s = new Random();
		int limit = N;
		byte[] buf;
		double fraction;
		String[] array = new String[MAX]; // to avoid uninitialization error
		String[] prevArray = array;
		int itemSetSize = 0;
		int numElem;
		for (int k = 0; k < L; k++){
			// choosing the size of item-set, instead of poisson distribution,
			// we will keep it fixed = I, instead of first generating itemsets
			// and then picking all itemsets of size I.
			//itemSetSize = poisson.nextInt();
			itemSetSize = I;
//			System.out.println("item set size = " + itemSetSize);
			if (k == 0){
				// choosing elements for first item set
				array = new String[itemSetSize];
				for (int l = 0; l < itemSetSize; l++){
					array[l] = items[s.nextInt(limit)];
				}
			}
			else { // k > 0
				// choosing elements for next item set
				// fraction determines how many elements to pick from previous item set
				fraction = exp.nextDouble();
				numElem = (int) (fraction * (double) itemSetSize);
//				System.out.println("numElem = " + numElem + ", itemSetSize =" + itemSetSize + ", k = " + k);
				array = new String[itemSetSize];
				numElem = Math.min(numElem, prevArray.length);
				
				for (int l = 0; l < itemSetSize; l++){
					if (l < numElem){
						array[l] = prevArray[l];
					}
					else {
						array[l] = items[s.nextInt(limit)];
					}
				}
			}
			prevArray = (String[]) array.clone();
			Arrays.sort(array);
						
			for (int l = 0; l < itemSetSize; l++){
				buf = array[l].getBytes();
				out.write(buf);
				if(l != itemSetSize - 1){
					out.write(comma.getBytes());
				}
				else {
					out.write(newline.getBytes());
				}		
			}
			if (itemSetSize == 0) k = k - 1;
		}
	}
	
	public static void createTransactions() throws IOException {
		System.out.println("Step 3 of 3: Transactions generation");
		System.out.println("=============================");
		System.out.println("Number of transactions = " + D);
		
		if(!transactionsFile.exists()){
			transactionsFile.createNewFile();
		}
		else {
			System.out.println("File " + transactionsFile.getAbsolutePath() + " exists, skipping Transaction generation step");	
			return;
		}

		String[][] itemSets = new String[L][];
		List<String> sList;
		int i = 0;
		int j = 0;
		
		Scanner scanner = new Scanner(itemSetsFile);
		while (scanner.hasNext()){
			String string = scanner.next();
			Scanner innerScan = new Scanner(string).useDelimiter(",");
			j = 0;
			sList = new ArrayList<String>();
		//	size = innerScan.nextInt();
			while(innerScan.hasNext()){
				sList.add(innerScan.next());
				j++;
			}
			itemSets[i] = new String[sList.size()];
			sList.toArray(itemSets[i]);
			i++;
		}
		// create a weight to associate per item set, based on exponential distribution with unit mean
		float[] weights = new float[L];
		float[] normWeights = new float[L];
		float[] corruption = new float[L];
		float[] weightedArray = new float[L];
		RandomEngine engine = new DRand();
		double mean = 1.0;
		float sum = 0.0f;
		float sumWeight = 0.0f;
		Exponential exp = new Exponential(mean, engine);
		double poissonMean = (double) T;
		Poisson poisson = new Poisson(poissonMean, engine);
		int trSize = 0;
		double normMean = 0.5;
		double variance = 0.1;
		double stDev = Math.sqrt(variance);
		Normal norm = new Normal(normMean, stDev, engine);
		Random s = new Random();
		Random itemDrop = new Random();
		Random itemSetDrop = new Random();
		byte[] buf;
		
		for (int l = 0; l < L; l++){
			weights[l] = (float) exp.nextDouble();
			sum += weights[l];
		}
		sumWeight = 0.0f;
		// normalize the weights
		for (int l = 0; l < L; l++){
			normWeights[l] = weights[l] / sum ;
			sumWeight += normWeights[l];
			weightedArray[l] = sumWeight;
		}
		// create a corruption level array
		for (int l = 0; l < L; l++){
			corruption[l] = (float) norm.nextDouble();
			if (corruption[l] < 0.0f){
				corruption[l] = 0.0f;
			}
			if (corruption[l] > 1.0f){
				corruption[l] = 1.0f;
			}
		}
				
		
		OutputStream out = new FileOutputStream(transactionsFile);
		System.out.println("Writing to file: " + transactionsFile.getAbsolutePath());
		
		float toss;
		int k;
		String[] itemSetSelected;
		int itemSetSize;
		List<String> trList;
		List<String> leftOver = new ArrayList<String>();

		for (int l = 0; l < D; l++){
			if( l % 1000000 == 0)
			System.out.println("transactions done = " + l + ", total = " + D);
			// determine size of the transaction
			trSize = poisson.nextInt();
			trList = new ArrayList<String>();
			if (leftOver.size() > 0){
				trList = leftOver;
				leftOver = new ArrayList<String>();
			}
			while (trList.size() < trSize){
				// choose an itemset to insert in a transaction by flipping an L-sided weighted coin 
				// as long as the number of items in a transaction < trSize
				toss = s.nextFloat();
				k = 0;
				while (k < L && weightedArray[k] <= toss){
					k++;
				}
				if (k != 0) k--;
				// drop items from itemset based on corruption level
				itemSetSelected = itemSets[k];
				itemSetSize = itemSetSelected.length;
				while (itemDrop.nextFloat() < corruption[k] && itemSetSize > 0){
					itemSetSize --;
				}
				if((trList.size()+itemSetSize > trSize) && (itemSetDrop.nextFloat() < HALF)) {
					for (int m = 0; m < itemSetSize; m++) {
						leftOver.add(itemSetSelected[m]);
					}
					break;
				}
				else {
					for (int m = 0; m < itemSetSize; m++) {
						if(!trList.contains(itemSetSelected[m])){
							trList.add(itemSetSelected[m]);
						}
					}
				}
			}
			Collections.sort(trList);
			for (int q = 0; q < trList.size(); q++){
				buf = trList.get(q).getBytes();
				out.write(buf);
				if(q != trList.size() - 1){
					out.write(comma.getBytes());
				}
				else {
					out.write(newline.getBytes());
				}		
			}
			if (trList.size() == 0) l--;
		}
	}
	
	public static void main(String[] args) throws IOException {

		System.out.println("CreateSelfJoinData started");
		// specify path where we want to put the generated data

		for(int i=0; i < args.length; ++i) {
	  	if ("-out".equals(args[i])) {
	  		String dir = new String(args[++i]);
	  		dataPath = new File(dir);
	  	} else if ("-N".equals(args[i])) {
	  		N = Integer.parseInt(args[++i]);
	  	} else if ("-L".equals(args[i])) {
	  		L = Integer.parseInt(args[++i]);
	  	} else if ("-D".equals(args[i])) {
	  		D = Integer.parseInt(args[++i]);
	  	} else if ("-T".equals(args[i])) {
	  		T = Integer.parseInt(args[++i]);
	  	} else if ("-I".equals(args[i])) {
	  		I = Integer.parseInt(args[++i]);
	  	} else if ("-B".equals(args[i])) {
	  		B = Integer.parseInt(args[++i]);
	  	} else {
	  		System.out.println("Invalid argument = " + args[i] + ", ignoring"); 		
	  	}
	  }
		//createItems();
		//createItemSets();
		// SelfJoin will operate on ItemSets output data
		createTransactions();

		
		
		System.out.println("CreateSelfJoinData finished");
	}
}
